/* commands.c */

#include "kb_wmouse.h"

/* controller_command() */
status_t controller_command(uint8 command) {
  status_t write_failed;
  
  write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
  if(write_failed) {
    dprintf(CONTROLLER_COMMAND_IBUF_NOT_EMPTY);
    return B_ERROR;
  }
  
  isa->write_io_8(PS2_CMD_REG, command);
  
  write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
  if(write_failed) {
    dprintf(CONTROLLER_COMMAND_IBUF_NOT_EMPTY);
    return B_ERROR;
  }
  
dprintf("kb_wmouse:controller_command(): success!\n");
  return B_OK;
}

/* controller_command_compare_result() */
status_t controller_command_compare_result(uint8 command,
                                           uint8 expected_result) {
  uint8 actual_result;
  status_t read_failed;
  
  do {
    if (controller_command(command) != B_OK) return B_ERROR;
  
    read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
    if(read_failed != B_OK) {
      dprintf(GET_DATA_OBUF_EMPTY);
      return B_ERROR;
    }
    actual_result = isa->read_io_8(PS2_DATA_REG);
    if(actual_result == expected_result) return B_OK;
    if(actual_result != 0xFE /* NAK */) return B_ERROR;
  }   while(actual_result != expected_result);
}

/* keyboard_command() */
status_t keyboard_command(uint8 command) {
  status_t write_failed, read_failed;
  uint8 result;
  int retries;
  
  for(retries = 10; retries != 0; --retries) {
    write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
    if(write_failed != B_OK) {
      dprintf(PUT_DATA_IBUF_FULL);
      dprintf(KB_CMD_ERROR, command, result);
      return B_ERROR;
    }
  
    isa->write_io_8(PS2_DATA_REG, command);
    read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
    if(read_failed != B_OK) {
      dprintf(GET_DATA_OBUF_EMPTY);
      dprintf(KB_CMD_ERROR, command, result);
      return -1;
    }
  
    result = isa->read_io_8(PS2_DATA_REG);
    if(result == 0xFA) return B_OK;
    if(result != 0xFE) {
      dprintf(KB_CMD_ERROR, command, result);
      return B_ERROR;
    }
    /* result was 0xFE (NAK), so keep trying */
  }
  return B_ERROR;
}

/* keyboard_command_with_data() */
status_t keyboard_command_with_data(uint8 command, uint8 data) {
  int32 retries = 0;
  uint8 result = 0;
  status_t write_failed, read_failed;
  
  for(retries = 10; retries != 0; --retries) {
    write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
    if(write_failed != B_OK) {
      dprintf(PUT_DATA_IBUF_FULL);
      dprintf(KB_CMD_WITH_DATA_ERROR, command, result);
      return B_ERROR;
    }
  
    isa->write_io_8(PS2_DATA_REG, command);
  
    write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
    if(write_failed != B_OK) {
      dprintf(PUT_DATA_IBUF_FULL);
      dprintf(KB_CMD_WITH_DATA_ERROR, command, result);
      return B_ERROR;
    }
  
    isa->write_io_8(PS2_DATA_REG, data);
  
    read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
    if(read_failed != B_OK) {
      dprintf(GET_DATA_OBUF_EMPTY);
      dprintf(KB_CMD_WITH_DATA_ERROR, command, result);
      return B_ERROR;
    }
    result = isa->read_io_8(PS2_DATA_REG);
    if(result == 0xFA) return B_OK;
    if(result != 0xFE) {
      dprintf(KB_CMD_WITH_DATA_ERROR, command, result);
      return B_ERROR;
    }
    /* NAK, so retry */
  }
  return B_ERROR;
}

/* change_command_bits() */
status_t change_command_bits(uint8 bit_mask_to_clear, uint8 bit_mask_to_set) {
  int tries;
  uint8 result;
  status_t write_failed;

dprintf("kb_wmouse:change_command_bits(0x%02x, 0x%02x): cmd_byte == 0x%02x\n", bit_mask_to_clear, bit_mask_to_set, cmd_byte);
cmd_byte = (cmd_byte & ~bit_mask_to_clear) | bit_mask_to_set;
dprintf("kb_wmouse:change_command_bits(): cmd_byte sent == 0x%02x\n", cmd_byte);
  for(tries = 0; tries <= 2; ++tries) {
    result = controller_command(0x60);  // KBD_CCMD_WRITE_MODE
    if(result != 0) {
//      if(tries > 2) return B_ERROR;
    } /*else break*/;
  
    write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
    if(write_failed != B_OK) {
      dprintf(PUT_DATA_IBUF_FULL);
      if(tries > 2) return B_ERROR;
    }
    
    isa->write_io_8(PS2_DATA_REG, cmd_byte);
    return B_OK;
  }
  return B_ERROR;
}

/* init_cmd_byte */
/* Returns what Linux calls the `controller mode bits'. */
status_t init_cmd_byte(void) {
  status_t read_failed, err;
  int tries;

  for(tries = 0; tries <= 2; ++tries) {
    /* flush output buffer */
    read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
    while(read_failed == B_OK) {
      dprintf(FLUSHED_MSG, isa->read_io_8(PS2_DATA_REG));
      read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
    }
  
    err = controller_command(0x20);
    if(err != B_OK) continue;  /* try again */
  
    read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
    if(read_failed != B_OK) {
      dprintf(GET_DATA_OBUF_EMPTY);
      continue;  /* try again */
    }
    cmd_byte = isa->read_io_8(PS2_DATA_REG);
dprintf("kb_wmouse:init_cmd_byte(): value is 0x%02x\n", cmd_byte);
    return cmd_byte;
  }
}

/* active_keyboard_command_with_data() */
status_t active_keyboard_command_with_data(uint8 command, uint8 data) {
  status_t err, write_failed;
  int tries;
  
  for(tries = 0; tries <= 2; ++tries) {
    waiting = 1;
    write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
    if(write_failed != B_OK) {
      dprintf(PUT_DATA_IBUF_FULL);
      continue;
    } else {
      isa->write_io_8(PS2_DATA_REG, command);
    }
    
    err = acquire_sem_etc(kb_wait_sem, 1,
                       B_CAN_INTERRUPT | B_RELATIVE_TIMEOUT,
                       1000000LL);
    if(err == B_INTERRUPTED) {
      waiting = 0;
      return err;
    }
    if(err != B_OK) continue;
    if(keyboardResult[0] != 0xFA) continue;
    waiting = 1;
    
    write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
    if(write_failed != B_OK) {
      dprintf(PUT_DATA_IBUF_FULL);
    } else {
      isa->write_io_8(PS2_DATA_REG, data);
    }
    if(write_failed) continue;
    
    err = acquire_sem_etc(kb_wait_sem, 1,
                          B_CAN_INTERRUPT | B_RELATIVE_TIMEOUT,
                          1000000LL);
    if (err == B_INTERRUPTED) {
      waiting = 0;
      return err;
    }
    if(err != B_OK) continue;
    if(keyboardResult[0] != 0xFA) continue;
    return B_OK;
  }
  waiting = 0;
  return err;
}

/* active_keyboard_command_get_results() */
status_t active_keyboard_command_get_results(uint8 command,
                                             void *dest,
                                             uint32 result_size) {
  int32 tries = 0;
  status_t write_failed, err;
  
  for(tries = 0; tries <= 2; ++tries) {
    waiting = result_size + 1;
  
    write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
    if(write_failed != B_OK) {
      dprintf(PUT_DATA_IBUF_FULL);
      continue;
    } else {
      isa->write_io_8(PS2_DATA_REG, command);
    }
  
    err = acquire_sem_etc(kb_wait_sem, 1,
                          B_CAN_INTERRUPT | B_RELATIVE_TIMEOUT,
                          1000000LL);
    if(err == B_INTERRUPTED) {
      waiting = 0;
      return err;
    }
    if(err != B_OK) continue;
  
    if(keyboardResult[result_size] != 0xFA) continue;
    memcpy(dest, keyboardResult, result_size);
    return 0;
  }  /* for */
}
